from ewah.operators.base import EWAHBaseOperator
from ewah.constants import EWAHConstants as EC

from ewah.hooks.base import EWAHBaseHook as BaseHook

import googlemaps


class EWAHGMapsOperator(EWAHBaseOperator):
    """Gets address data from a custom SQL fired against the DWH.

    Params:
        - address_sql - a (templated) SQL statement that fetches the data
            as required; only one column called 'address' will be respected.
            Any additional columns will be ignored, and failure to provide
            such a column will cause failure of the operator.
    """

    _NAMES = ["gmaps", "google_maps", "googlemaps"]

    _ACCEPTED_EXTRACT_STRATEGIES = {
        EC.ES_FULL_REFRESH: True,
        EC.ES_INCREMENTAL: True,  # use templating for incremental usecases
    }

    def __init__(self, address_sql, *args, **kwargs):
        self.template_fields.add("address_sql")
        kwargs["primary_key"] = "address"
        super().__init__(*args, **kwargs)

        self.address_sql = address_sql

    def ewah_execute(self, context):
        dwh_hook = self.upload_hook

        addresses = dwh_hook.execute_and_return_result(
            sql=self.address_sql,
            return_dict=True,
        )

        api_key = self.source_conn.password

        client = googlemaps.Client(key=api_key)

        data = []
        while addresses:
            try:
                address = addresses.pop(0)["address"]
            except:
                raise Exception("Your SQL did not provide an address column!")

            data += [
                {
                    "address": address,
                    "geocode_result": client.geocode(address),
                }
            ]

        self.upload_data(data)



from ewah.constants import EWAHConstants as EC
from ewah.hooks.base import EWAHBaseHook

from datetime import datetime, date
from typing import Optional, Dict, List

import json
import requests


class EWAHLinkedInHook(EWAHBaseHook):
    _ATTR_RELABEL: {}

    conn_name_attr = "ewah_linkedin_conn_id"
    default_conn_name = "ewah_linkedin_default"
    conn_type = "ewah_linkedin"
    hook_name = "EWAH LinkedIn Connection"

    @staticmethod
    def get_ui_field_behaviour() -> dict:
        return {
            "hidden_fields": ["extra", "host", "port", "login", "schema"],
            "relabeling": {
                "password": "Access Token",
            },
        }

    _AVAILABLE_FIELDS_ANALYTICS = {
        "actionClicks",
        "adUnitClicks",
        "approximateUniqueImpressions",
        "cardClicks",
        "cardImpressions",
        "clicks",
        "commentLikes",
        "comments",
        "companyPageClicks",
        "conversionValueInLocalCurrency",
        "costInLocalCurrency",
        "costInUsd",
        "dateRange",
        "externalWebsiteConversions",
        "externalWebsitePostClickConversions",
        "externalWebsitePostViewConversions",
        "follows",
        "fullScreenPlays",
        "impressions",
        "landingPageClicks",
        "leadGenerationMailContactInfoShares",
        "leadGenerationMailInterestedClicks",
        "likes",
        "oneClickLeadFormOpens",
        "oneClickLeads",
        "opens",
        "otherEngagements",
        "pivot",
        "pivotValue",
        "pivotValues",
        "reactions",
        "sends",
        "shares",
        "textUrlClicks",
        "totalEngagements",
        "videoCompletions",
        "videoFirstQuartileCompletions",
        "videoMidpointCompletions",
        "videoStarts",
        "videoThirdQuartileCompletions",
        "videoViews",
        "viralCardClicks",
        "viralCardImpressions",
        "viralClicks",
        "viralCommentLikes",
        "viralComments",
        "viralCompanyPageClicks",
        "viralExternalWebsiteConversions",
        "viralExternalWebsitePostClickConversions",
        "viralExternalWebsitePostViewConversions",
        "viralFollows",
        "viralFullScreenPlays",
        "viralImpressions",
        "viralLandingPageClicks",
        "viralLikes",
        "viralOneClickLeadFormOpens",
        "viralOneClickLeads",
        "viralOtherEngagements",
        "viralReactions",
        "viralShares",
        "viralTotalEngagements",
        "viralVideoCompletions",
        "viralVideoFirstQuartileCompletions",
        "viralVideoMidpointCompletions",
        "viralVideoStarts",
        "viralVideoThirdQuartileCompletions",
        "viralVideoViews",
    }

    _REQUIRED_FIELDS_ANALYTICS = {
        "dateRange",
        "pivotValue",
    }

    _AVAILABLE_OBJECTS = {"Accounts", "Campaigns", "Creatives", "CampaignGroups"}

    @staticmethod
    def get_cleaner_callables():
        # dateRange madness is fixed here
        def cast_date_range(row):
            date_range = row.pop("dateRange", None)
            if not date_range:
                return row
            date_start = date_range.pop("start")
            date_end = date_range.pop("end")

            row["date_start"] = date(
                date_start["year"], date_start["month"], date_start["day"]
            )
            row["date_end"] = date(date_end["year"], date_end["month"], date_end["day"])

            return row

        return [cast_date_range]

    @classmethod
    def validate_fields_list(cls, fields_list: List[str]) -> bool:
        assert isinstance(fields_list, list), "Must supply a list of fields!"
        for field in cls._REQUIRED_FIELDS_ANALYTICS:
            assert field in fields_list, f"Must include the field {field}!"
        assert len(fields_list) <= 20, "Can only have a maximum of 20 fields!"
        for field in fields_list:
            assert (
                field in cls._AVAILABLE_FIELDS_ANALYTICS
            ), "Field {0} not part of allowed fields:\n\n{1}".format(
                field, "\n\t".join(cls._AVAILABLE_FIELDS_ANALYTICS)
            )
        return True

    @property
    def call_headers(self) -> Dict[str, str]:
        return {
            "Authorization": "Bearer {0}".format(self.conn.password),
            "Accept": "application/json",
        }

    def get_account_list(self, return_only_ids=True):
        return [account["id"] for account in self.get_object_data("Accounts")]

    def get_analytics_for_account(
        self, account_id, pivot, date_start, date_end, fields
    ):
        # Note that pagination is not supported for this endpoint
        self.validate_fields_list(fields)
        url = "https://api.linkedin.com/v2/adAnalyticsV2"
        self.log.info("Pulling analytics data for account {0}".format(account_id))
        params = {
            "q": "analytics",
            "pivot": pivot,
            "dateRange.start.day": date_start.day,
            "dateRange.start.month": date_start.month,
            "dateRange.start.year": date_start.year,
            "dateRange.end.day": date_end.day,
            "dateRange.end.month": date_end.month,
            "dateRange.end.year": date_end.year,
            "timeGranularity": "DAILY",
            "accounts[0]": "urn:li:sponsoredAccount:{0}".format(account_id),
            "fields": ",".join(fields),
        }

        response = requests.get(url, params=params, headers=self.call_headers)
        assert response.status_code == 200, response.text
        return response.json()["elements"]

    def get_analytics_in_batches(
        self, pivot, date_start, date_end, fields, account_ids=None
    ):
        if not account_ids:
            account_ids = self.get_account_list()

        for account_id in account_ids:
            yield self.get_analytics_for_account(
                account_id, pivot, date_start, date_end, fields
            )

    def get_object_data(self, object_name):
        url = "https://api.linkedin.com/v2/ad{0}V2".format(object_name)
        params = {
            "q": "search",
            "start": 0,
            "count": 100,  # 100 is the max value for count
        }
        data = []
        while True:
            response = requests.get(url, params=params, headers=self.call_headers)
            assert response.status_code == 200, response.text
            elements = response.json()["elements"]
            if elements:
                data += elements
                if len(elements) < params["count"]:
                    break
                params["start"] = params["start"] + params["count"]
            else:
                break
        return data



from ewah.operators.sql_base import EWAHSQLBaseOperator
from ewah.hooks.mysql import EWAHMySQLHook


class EWAHMySQLOperator(EWAHSQLBaseOperator):
    _NAMES = ["mysql"]

    _SQL_BASE = "SELECT\n{columns}\nFROM `{schema}`.`{table}`"
    _SQL_BASE_SELECT = "SELECT * FROM (\n\n{select_sql}\n\n) t\n\nWHERE {{0}}"
    _SQL_COLUMN_QUOTE = "`"
    _SQL_PARAMS = "%({0})s"

    _CONN_TYPE = EWAHMySQLHook.conn_type



from ewah.hooks.base import EWAHBaseHook

from typing import Dict, List, Any, Optional

import stripe


class EWAHStripeHook(EWAHBaseHook):
    _ATTR_RELABEL = {
        "api_key": "password",
    }

    conn_name_attr = "ewah_stripe_conn_id"
    default_conn_name = "ewah_stripe_default"
    conn_type = "ewah_stripe"
    hook_name = "EWAH Stripe Connection"

    _listable = stripe.api_resources.abstract.ListableAPIResource
    _singleton = stripe.api_resources.abstract.SingletonAPIResource

    @staticmethod
    def get_ui_field_behaviour() -> Dict:
        return {
            "hidden_fields": ["port", "schema", "extra", "host", "login"],
            "relabeling": {
                "password": "API Key",
            },
        }

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # authenticate using API key
        stripe.api_key = self.conn.api_key

    def get_resource(self, resource: str):
        """Utility function to retrieve a particular resource.

        :param resource: path of the resource. E.g. "issuing.Transaction" to get
            stripe.api_resources.issuing.Transaction.
        :return: Stripe API Resource.
        """
        result = stripe.api_resources
        for subresource in resource.split("."):
            result = getattr(result, subresource)
        return result

    def get_data_in_batches(
        self,
        resource: str,
        expand: Optional[List[str]] = None,
        batch_size: int = 10000,
    ) -> List[Dict[str, Any]]:
        # get data
        data = []
        resource = self.get_resource(resource=resource)
        if issubclass(resource, self._listable):
            all_items = resource.list(limit=100, expand=expand)
            for item in all_items.auto_paging_iter():
                data += [item.to_dict_recursive()]
                if len(data) >= batch_size:
                    yield data
                    data = []
            if data:  # Last batch
                yield data
        elif issubclass(resource, self._singleton):
            data = [resource.retrieve(expand=expand).to_dict_recursive()]
            # singletons have no id, but we need it as PK -> set it manually:
            data[0]["id"] = 1
            assert len(data) == 1, "Not a singleton!"
            yield data
        else:
            raise Exception("Invalid resource!")



from ewah.hooks.base import EWAHBaseHook
from pyairtable import Table


class EWAHAirtableHook(EWAHBaseHook):
    _ATTR_RELABEL = {
        "api_key": "password",
    }

    conn_name_attr = "ewah_airtable_conn_id"
    default_conn_name = "ewah_airtable_default"
    conn_type = "ewah_airtable"
    hook_name = "EWAH Airtable Connection"

    @staticmethod
    def get_ui_field_behaviour():
        return {
            "hidden_fields": ["port", "schema", "extra", "host", "login"],
            "relabeling": {"password": "Airtable API Key"},
        }

    def get_table_data(self, base_id, table_id):
        data = Table(self.conn.api_key, base_id, table_id).all()
        for datum in data:
            # All fields are contained in a dictionary
            # -> Move it into the first level
            datum.update(datum.pop("fields"))
        return data



from ewah.hooks.base import EWAHBaseHook

import requests


class EWAHInfigoHook(EWAHBaseHook):
    _ATTR_RELABEL = {
        "api_token": "password",
        "base_url": "host",
    }

    conn_name_attr = "ewah_infigo_conn_id"
    default_conn_name = "ewah_infigo_default"
    conn_type = "ewah_infigo"
    hook_name = "EWAH Infigo Connection"

    @staticmethod
    def get_ui_field_behaviour():
        return {
            "hidden_fields": ["port", "schema", "extra", "login"],
            "relabeling": {"password": "API Token", "host": "Infigo Base URL"},
        }

    def make_api_request(self, url):
        response = requests.get(
            url, headers={"Authorization": "Basic {0}".format(self.conn.api_token)}
        )
        assert response.status_code == 200, "Error - Response {0}: {1}".format(
            response.status_code, response.text
        )
        return response.json()

    def get_orders(self, base_url):
        orders = self.make_api_request(base_url + "Order/List")
        data = []
        while orders:
            order = self.make_api_request(
                base_url + "Order/Detail/" + str(orders.pop(0))
            )
            order["OrderLineItems"] = [
                self.make_api_request(base_url + "OrderlineItem/Get/" + str(line_item))
                for line_item in order.get("OrderLineItems", [])
            ]
            data.append(order)
        return data

    def get_customers(self, base_url):
        customers = self.make_api_request(base_url + "customer")
        data = []
        while customers:
            data.append(
                self.make_api_request(
                    base_url + "Customer/Get/" + str(customers.pop(0))
                )
            )
        return data

    def get_printlocations(self, base_url):
        return self.make_api_request(base_url + "PrintLocation")

    def get_data(self, resource):
        base_url = self.conn.base_url
        if not base_url.endswith("/"):
            base_url += "/"
        if not base_url.endswith("services/api/"):
            base_url += "services/api/"
        resource = resource.lower()

        if resource in ["orders", "orders"]:
            return self.get_orders(base_url)
        elif resource in ["customer", "customers"]:
            return self.get_customers(base_url)
        elif resource in ["printlocation", "printlocations"]:
            return self.get_printlocations(base_url)
        else:
            raise Exception(
                "Invalid resource! may only fetch orders, customers, or printlocations!"
            )



from ewah.constants import EWAHConstants as EC
from ewah.hooks.google_cloud_storage import EWAHGoogleCloudStorageHook
from ewah.operators.base import EWAHBaseOperator


class EWAHGoogleCloudStorageOperator(EWAHBaseOperator):
    _NAMES = ["gcs", "google_cloud_storage"]

    _ACCEPTED_EXTRACT_STRATEGIES = {
        EC.ES_FULL_REFRESH: True,
        EC.ES_INCREMENTAL: False,
        EC.ES_SUBSEQUENT: True,
    }

    _CONN_TYPE = EWAHGoogleCloudStorageHook.conn_type

    def __init__(self, bucket_name, prefix=None, data_after=None, *args, **kwargs):
        self.bucket_name = bucket_name
        self.data_after = data_after
        self.prefix = prefix
        kwargs["subsequent_field"] = "blob_modified_at"
        kwargs["primary_key"] = "id"
        super().__init__(*args, **kwargs)

    def ewah_execute(self, context):
        if (
            self.extract_strategy == EC.ES_SUBSEQUENT
            and self.test_if_target_table_exists()
        ):
            data_from = self.get_max_value_of_column(self.subsequent_field)
        else:
            data_from = self.data_after or self.data_from

        for batch in self.source_hook.get_data_in_batches(
            self.bucket_name, data_after=data_from, prefix=self.prefix
        ):
            self.upload_data(batch)



from ewah.hooks.personio import EWAHPersonioHook

from ewah.operators.base import EWAHBaseOperator
from ewah.constants import EWAHConstants as EC


class EWAHPersonioOperator(EWAHBaseOperator):
    _NAMES = ["personio"]

    _ACCEPTED_EXTRACT_STRATEGIES = {
        EC.ES_FULL_REFRESH: True,
        EC.ES_INCREMENTAL: False,
        EC.ES_SUBSEQUENT: True,
    }

    _CONN_TYPE = EWAHPersonioHook.conn_type

    def __init__(self, resource=None, *args, **kwargs):
        resource = resource or kwargs.get("target_table_name", "").lower()
        assert EWAHPersonioHook.validate_resource(resource)
        if kwargs.get("extract_strategy") == EC.ES_SUBSEQUENT:
            assert (
                resource == "attendances"
            ), "Only attendances work with subsequent loading!"
            kwargs["subsequent_field"] = "updated_at"
            kwargs["primary_key"] = "id"

        super().__init__(*args, **kwargs)

        self.resource = resource

    def ewah_execute(self, context):
        if (
            self.extract_strategy == EC.ES_SUBSEQUENT
            and self.test_if_target_table_exists()
        ):
            data_from = self.get_max_value_of_column(self.subsequent_field)
        else:
            data_from = None
        for batch in self.source_hook.get_data_in_batches(
            resource=self.resource,
            data_from=data_from,
        ):
            self.upload_data(batch)
